-- @author Stephan Hennig <stephanhennig@arcor.de>
-- @release v0.2, 2010-02-27
-- @copyright 2009, 2010, Stephan Hennig
-- trie.lua, auxillary trie related functions
--
--Copyright (C) 2009, 2010 Stephan Hennig
--
--This file is part of the pattern statistics bundle.
--
--The pattern statistics bundle is free software; you can redistribute
--it and/or modify it under the terms of the GNU General Public License
--as published by the Free Software Foundation; either version 3
--of the License, or (at your option) any later version.
--
--The pattern statistics bundle is distributed in the hope that it will
--be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
--of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.
--
--You should have received a copy of the GNU General Public License
--along with this program; if not, see <http://www.gnu.org/licenses/>.


--- Create a new empty trie node.
-- An empty node is a table { next = {}, value = nil }.
-- @return An empty trie node.
function newTrieNode()
  return {
      next = {},
      value = nil
    }
end


--- Insert into a trie a plain string (not taking sub-strings into account)
-- with an associated value.  An existing value is replaced.
-- @param trie  The trie object to insert the string into.
-- @param s  The string to insert.
-- @param value  A non-nil value associated with the string.
function insertStringSingle(trie, s, value)
  local node = trie
  for ch in string.gmatch(s, ".") do
    if node.next[ch] == nil then
      node.next[ch] = newTrieNode()
    end
    node = node.next[ch]
  end
--~   if node.value ~= nil then
--~     print("Wort " .. s .. " bereits enthalten.")
--~   end
  node.value = value
end


--- Insert into a trie all sub-strings of a string starting at position 1
-- (there are n = len(s) of them).
-- Example:  insert(t, "Hello", value) inserts strings "H", "He", "Hel",
-- "Hell", "Hello" with associated values value[1], value[2], ..., value[5]
-- into trie t.  Any existing values are replaced.
-- @param trie  The trie object to insert the string into.
-- @param s  The string to insert.
-- @param value  A table of length len(s), containing non-nil values associated
-- with the sub-strings.  Alternatively, value can be any non-table expression.
-- This value is then associated with all sub-strings.
function insertStringCummulative(trie, s, value)
  local node = trie
  local vtable = type(value) == "table"
  local i = 0
  for ch in string.gmatch(s, ".") do
    i = i + 1
    if node.next[ch] == nil then
      node.next[ch] = newTrieNode()
    end
    node = node.next[ch]
    node.value = vtable and value[i] or value
  end
end


--- Insert into a trie all sub-strings of a string (there are
-- n = len(s) * ( len(s) + 1 ) / 2 of them).
-- Example:  insert(t, "Hello") inserts strings "H", "He", "Hel", "Hell",
-- "Hello" with associated values value[1][1], value[1][2], ..., value[1][5],
-- strings "e", "el", "ell", "ello" with associated values value[2][1],
-- value[2][2], ..., value[2][4], etc into trie t.  Any existing values will
-- be replaced.  If there are identical sub-strings, the final associated
-- value is not determined.
-- @param trie  The trie object to insert the string into.
-- @param s  The string to insert.
-- @param value  A triangular table of length len(s), containing tables of
-- non-nil values associated with the sub-strings.  Alternatively, value can
-- be a non-table type expression.  This value is then associated with all
-- sub-strings.
function insertStringFull(trie, s, value)
  local vtable = type(value) == "table"
  for i = 1, string.len(s) do
    insertStringCummulative(trie, string.sub(s, i, -1), vtable and value[i] or value)
  end
end


--- Insert into a trie a list of plain strings.
-- @param trie  A trie to insert strings into.
-- @param list  A list with strings.  All strings are associated with a
-- value true.
function insertListSingle(trie, list)
  for _, s in pairs(list) do
    insertStringSingle(trie, s, true)
  end
end


--- Insert into a trie all sub-strings starting at position 1 of a list
-- of strings.
-- @param trie  A trie to insert strings into.
-- @param list  A list with strings.  All sub-strings are associated with a
-- value true.
function insertListCummulative(trie, list)
  for _, s in pairs(list) do
    insertStringCummulative(trie, s, true)
  end
end


--- Insert into a trie all sub-strings of a list of strings.
-- @param trie  A trie to insert strings into.
-- @param list  A list with strings.  All sub-strings are associated with a
-- value true.
function insertListFull(trie, list)
  for _, s in pairs(list) do
    insertStringFull(trie, s, true)
  end
end


--- Insert into a trie all plain lines of a file.
-- @param trie  A trie to insert lines into.
-- @param file  File to read from.  All strings are associated with a
-- value true.
function insertFileSingle(trie, file)
  io.input(file)
  for line in io.lines() do
    insertStringSingle(trie, string.lower(line), true)
  end
end


--- Insert into a trie all sub-strings starting at position 1 of the lines
-- read from a file.
-- @param trie  A trie to insert strings into.
-- @param file  File to read from.  All strings are associated with a
-- value true.
function insertFileCummulative(trie, file)
  io.input(file)
  for line in io.lines() do
    insertStringCummulative(trie, string.lower(line), true)
  end
end


--- Insert into a trie all sub-strings of the lines read from a file.
-- @param trie  A trie to insert strings into.
-- @param file  File to read from.  All strings are associated with a
-- value true.
function insertFileFull(trie, file)
  io.input(file)
  for line in io.lines() do
    insertStringFull(trie, string.lower(line), true)
  end
end


--- Search for a plain string in a trie.
-- @param trie  The trie object.
-- @param s  String to search for.
-- @return If the strings is found, return the associated value.  If the
-- string could not be found, return nil.
function matchStringSingle(trie, s)
  local node = trie
  local i = 0
  for ch in string.gmatch(s, ".") do
    node = node.next[ch]
    if not node then break end
    i = i + 1
  end
  return i == string.len(s) and node.value or nil
end


--- Match all sub-strings of a string starting at position 1 against a trie.
-- @param trie  The trie object.
-- @param s  String to search for.
-- @return If sub-strings have been identified, return a list with values
-- associated with the matching strings.  If no matching sub-strings
-- have been identified, return nil.
function matchStringCummulative(trie, s)
  local node = trie
  local values = {}
  local i = 0
  for ch in string.gmatch(s, ".") do
    node = node.next[ch]
    if not node then break end
    if node.value ~= nil then
      i = i + 1
      values[i] = node.value
    end
  end
  return i > 0 and values or nil
end


--- Match all sub-strings of a string against a trie.
-- @param trie  The trie object.
-- @param s  String to search for.
-- @return If sub-strings have been identified, return a list with values
-- associated with the matching strings.  If no matching sub-strings
-- have been identified, return nil.
function matchStringFull(trie, s)
  local value = {}
  local j = 0
  for i = 1, string.len(s) do
    local val = matchStringCummulative(trie, string.sub(s, i, -1))
    if val then
      for _, v in pairs(val) do
        j = j + 1
        value[j] = v
      end
    end
  end
  return j > 0 and value or nil
end


--- Print all strings stored in a trie.
-- An exclamation mark indicates identified strings.
-- @see showTrie__
-- @param trie  The trie object to show.
-- @param compact  boolean flag: Print strings in compact format?
function showTrie(trie, compact)
  showTrie__(trie, "", compact)
end


--- This function recursively advances through the trie.
-- It prints all characters valid at a node.
-- An underscore indicates the end of an identified string.
-- @usage Internal function.
-- @see showTrie
-- @param node  A trie object.
-- @param s  String of characters preceeding the current node.
-- @param compact  boolean flag: Print patterns in compact format?
function showTrie__(node, s, compact)
  local found = ""
  if node.value ~= nil then
    found = "_"
  end
  io.write(found)
  local count = 0
  for ch, next in pairs(node.next) do
    count = count + 1
    if count > 1 then
      if compact then
        io.write("\n", string.rep(" ", string.len(s) + string.len(found)))
      else
        io.write("\n", s, found)
      end
    end
    io.write(ch)
    showTrie__(next, s..found..ch, compact)
  end
end
