-- @author Stephan Hennig <stephanhennig@arcor.de>
-- @release v0.2, 2010-02-27
-- @copyright 2009, 2010, Stephan Hennig
-- words.lua, auxillary word related functions
--
--Copyright (C) 2009, 2010 Stephan Hennig
--
--This file is part of the pattern statistics bundle.
--
--The pattern statistics bundle is free software; you can redistribute
--it and/or modify it under the terms of the GNU General Public License
--as published by the Free Software Foundation; either version 3
--of the License, or (at your option) any later version.
--
--The pattern statistics bundle is distributed in the hope that it will
--be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
--of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.
--
--You should have received a copy of the GNU General Public License
--along with this program; if not, see <http://www.gnu.org/licenses/>.


--- Decompose a word into its constituent patterns.
-- The word is padded with "." and converted to lower-case before
-- matching it to the trie.
-- @param trie  The trie to be used for decomposition.
-- @param word  Word to decompose.
-- @return A table whose values are the asssociated values of identified
-- patterns.  There's an entry for every single match of a pattern.
function decomposeWord(trie, word)
  local w = "."..string.lower(word).."."
  local patterns = {}
  local ipatterns = 0
  for pos = 1, string.len(w) do
    local subw = string.sub(w, pos, -1)
    local values = matchStringCummulative(trie, subw)
    if values then
      for _, v in pairs(values) do
        ipatterns = ipatterns + 1
        patterns[ipatterns] = v
      end
    end
  end
  return patterns
end

--- Decompose a word into its constituent patterns (alternative
-- implementation).
-- Considers all characters in a string only once.  It manages a
-- queue of active tries into which the same character is fed into in
-- parallel.  This one is a bit faster than decomposeWord(); upto twice
-- as fast, at the expense of memory, if the line 'queue[qmin] = nil'
-- is omitted.
-- @see decomposeWord
function decomposeWord2(trie, word)
  local w = "."..string.lower(word).."."
  local patterns = {}
  local ipatterns = 0
  local queue = {}
  local qmin = 1
  local qmax = 0
  for ch in string.gmatch(w, ".") do
    qmax = qmax + 1
    queue[qmax] = trie
--~     print(qmin, qmax, qmax - qmin)
    local lower = qmin
    local upper = qmax
    for i = lower, upper do
      local node = queue[qmin].next[ch]
      queue[qmin] = nil
      qmin = qmin + 1
      if node then
        qmax = qmax + 1
        queue[qmax] = node
        if node.value ~= nil then
          ipatterns = ipatterns + 1
          patterns[ipatterns] = node.value
        end
      end
    end
  end
  return patterns
end


--- Output statistics on patterns matching a specific word.
-- The output is of the form word, total number of patterns matching the word
-- (taking multiple matches of a pattern into account), the hyphenation
-- patterns matching the word.  Patterns matching multiple times are appended
-- by a colon followed by the number of matches of the patterns.  All output
-- is placed on one line per word.
-- @param word  String containing the word in question.
-- @param values  A list of values associated with a pattern as provided by
-- decomposeWord() or decomposeWord2().
-- @see decomposeWord, decomposeWord2
function showDecomposedWord(word, values)
  local count = 0
  local patterns = {}
  for _, v in pairs(values) do
    count = count + 1
    patterns[v] = (patterns[v] or 0) + 1
  end
  io.write(word, " ", count)
  for k, v in pairs(patterns) do
    io.write(" ", k)
    if v > 1 then
      io.write(":", v)
    end
  end
  io.write("\n")
end
